import webbrowser

from fluxwallet.transactions import BaseTransaction
from fluxwallet.wallet import WalletTransaction
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Button, Label, RichLog


class ExplorerLabel(Label):
    def action_open_explorer_link(self, txid: str):
        base = "https://explorer.runonflux.io/tx"
        webbrowser.open(f"{base}/{txid}")


class TxOverlay(ModalScreen):
    DEFAULT_CSS = """
    TxOverlay {
        align: center middle;
    }

    TxOverlay > Container {
        padding: 0 1;
        width: 90%;
        height: 90%;
        border: thick $background 80%;
        background: $surface;
    }
    TxOverlay > Container > Horizontal {
        border: tall $primary;
        height: 5;
        margin-bottom: 1;
        background: $primary-background;
    }

    #txid {
        margin-left: 2;
        height: 3;
        # height: 1fr;
        # width: 1fr;
        content-align: center middle;
        padding-left: 1;
        padding-right: 1;
    }

    #exit {
        margin-right: 2;
        min-width: 5;
        width: 5;
    }
    #button_container {
        align: right middle;
    }
    TxOverlay > Container > RichLog {
        padding: 2;
        border: tall $primary;
    }
    """

    BINDINGS = [
        ("ctrl+t", "app.toggle_dark", "Toggle Dark mode"),
        Binding("ctrl+c,ctrl+q", "app.quit", "Quit"),
    ]

    def __init__(self, tx: WalletTransaction):
        self.wt = tx
        super().__init__()

    def compose(self) -> ComposeResult:
        label = f"View on explorer: [@click=open_explorer_link('{self.wt.tx.txid}')]{self.wt.tx.txid}[/]"

        yield Container(
            Horizontal(
                ExplorerLabel(label, id="txid"),
                Container(
                    Button("X", variant="error", id="exit"), id="button_container"
                ),
            ),
            RichLog(auto_scroll=False),
        )

    def on_mount(self):
        textlog = self.query_one("RichLog", RichLog)
        textlog.write(self.wt.tx.as_dict())
        textlog.write(" ")
        textlog.focus()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss()



from typing import Iterable

from textual import on
from textual.app import App, ComposeResult
from textual.containers import Center, Container, Horizontal
from textual.messages import Message
from textual.reactive import reactive
from textual.validation import Function, Length, Number
from textual.widget import Widget
from textual.widgets import Button, Input, Label, Select, Static

# from .top_bar import TopBar

# address_book = {
#     "gravywallet": "t1XvGeQCfYMhfagYb3GmbKRajNEjpPRZHkB",
#     "Anwen": "t1CvGeCCfYMhfagYAba4Re3ajNEjpPRZJ74",
# }


class Send(Widget):
    DEFAULT_CSS = """
    # Send {
    #     height: 100%;
    #     width: 1fr;
    #     min-width: 60;
    #     # align: center middle;
    #     outline: tall $primary;
    #     margin-top: 1;
    #     margin-bottom: 1;
    #     background: $panel;
    #     layers: default notification;
    # }
    # Send>Input {
    #     margin: 1;
    # }
    # Send>Horizontal {
    #     margin-top: 1;
    #     margin-left: 1;
    #     height: 5;
    # }
    # Send>Horizontal>Select {
    #     width: 1fr;
    #     max-height: 3;
    #     margin-left: 2;
    # }
    # Send>Input {
    #     border: tall $primary;
    # }
    # Send>Horizontal>Input {
    #     width: 2fr;
    #     border: tall $primary;
    #     # margin: 1;
    # }
    # Send>Center {
    #     margin-left: 1;
    #     margin-top: 1;
    # }
    # #save {
    #     margin: 0;
    #     padding: 0;
    #     height: auto;
    #     min-width: 0;
    #     width: auto;
    # }
    # #save:focus {
    #     text-style: none;
    #     outline-top: tall black;
    #     background: black 80%;
    # }
    # #buttons {
    #     margin-left: 1;
    #     margin-top: 1;
    #     align: center middle;
    # }
    # #all_funds {
    #     margin: 0;
    #     padding-right: 1;
    #     margin-left: 1;
    #     outline-right: tall black;
    #     min-width: 0;
    # }
    # #all_funds:focus {
    #     text-style: none;
    #     outline-top: tall black;
    #     background: black 80%;
    # }
    # Send>Horizontal>Button {
    #     margin: 1;

    # }
    # Send>Horizontal>Select>SelectCurrent {
    #     border: tall $primary;
    #     max-height: 3;
    # }
    # Send>Horizontal>Select>SelectCurrent Static#label {
    #    overflow: hidden;
    # }
    """

    sendto_address = reactive("")
    sendto_amount = reactive(0.0)
    send_enabled = reactive(False)
    save_enabled = reactive(False)

    class SendTxRequested(Message):
        def __init__(self, address: str, amount: float, message: str):
            self.address = address
            self.amount = amount
            self.message = message
            super().__init__()

    class MaxAmountRequested(Message):
        ...

    class AddressBookUpdateRequested(Message):
        def __init__(self, address: str):
            self.address = address
            super().__init__()

    def __init__(self, address_book: dict[str, str] = {}):
        self.address_book = address_book
        self.disabled_overlay = Static("Coming soon", id="disabled_send_banner")
        self.disabled_mounted = False
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Label("Send coins", classes="top")
        yield Horizontal(
            Input(
                placeholder="Address",
                id="address",
                validators=[
                    Function(function=self.validate_address_input),
                ],
            ),
            Button("\U0001F4BE", id="save"),
            Select(
                id="address_book",
                options=[(x, x) for x in self.address_book],
                prompt="Address book",
            ),
        )
        yield Horizontal(
            Input(
                placeholder="Amount",
                validators=[
                    Function(function=self.validate_amount_input),
                ],
                id="amount",
            ),
            Button("Max", id="all_funds"),
        )
        yield Input(placeholder="Message", id="message")
        yield Horizontal(
            Button("Clear", variant="warning", id="clear"),
            Button("Send", disabled=True, id="send"),
            id="buttons",
        )
        yield Center(id="disabled_container")

    def validate_address_input(self, address: str) -> bool:
        if address == "":
            return True

        if address.startswith(("t1", "t3")) and len(address) == 35:
            self.sendto_address = address
            return True
        return False

    def validate_amount_input(self, amount: float | str) -> bool:
        if amount == "":
            return True

        validator = Number(minimum=0)
        return not bool(validator.validate(amount).failures)

    def on_mount(self):
        self.query_one("#address", Input).focus()

    def update_address_book(self, selected: str, address_book: dict[str, str]):
        self.address_book = address_book
        options_list = sorted([(x, x) for x in self.address_book.keys()])

        address_book_dom = self.query_one("#address_book", Select)
        address_book_dom.set_options(options_list)

        if selected:
            address_book_dom.value = selected

    def mount_disabled(self) -> None:
        if not self.disabled_mounted:
            self.disabled_mounted = True
            self.disabled = True
            container = self.query_one("#disabled_container", Center)
            container.mount(self.disabled_overlay)

    def unmount_disabled(self) -> None:
        if self.disabled_mounted:
            self.disabled_mounted = False
            self.disabled = False
            self.disabled_overlay.remove()

    @on(Input.Changed, "#address")
    def address_input_changed(self, event: Input.Changed) -> None:
        event.stop()

        if not event.validation_result.failures:
            self.sendto_address = event.value
        else:
            self.sendto_address = ""

    @on(Input.Changed, "#amount")
    def amount_input_changed(self, event: Input.Changed) -> None:
        event.stop()

        if event.value == "":
            self.sendto_amount = 0.0
            return

        if not event.validation_result.failures:
            self.sendto_amount = float(event.value)
        else:
            self.sendto_amount = 0.0

    @on(Select.Changed, "#address_book")
    def address_book_select_changed(self, event: Select.Changed) -> None:
        event.stop()

        if event.value:
            self.query_one("#address", Input).value = self.address_book[
                str(event.value)
            ]

    @on(Button.Pressed, "#send")
    def submit_transaction(self):
        message = self.query_one("#message", Input).value
        self.post_message(
            self.SendTxRequested(self.sendto_address, self.sendto_amount, message)
        )
        # self.clear_fields()
        # self.query_one("#address", Input).focus()

    @on(Button.Pressed, "#all_funds")
    def set_max_amount(self):
        self.post_message(self.MaxAmountRequested())

    @on(Button.Pressed, "#clear")
    def clear_fields(self):
        self.sendto_address = ""
        self.sendto_amount = 0.0

        inputs: Iterable[Input] = self.query("Input")
        for input in inputs:
            input.value = ""

        self.query_one("#address_book", Select).value = None
        self.query_one("#address", Input).focus()

    @on(Button.Pressed, "#save")
    def save_address(self):
        address = self.query_one("#address", Input).value
        self.post_message(self.AddressBookUpdateRequested(address))

    def compute_send_enabled(self) -> bool:
        return True if self.sendto_address and self.sendto_amount else False

    def compute_save_enabled(self) -> bool:
        return bool(self.sendto_address)

    def watch_send_enabled(self, new_value: bool) -> None:
        button = self.query_one("#send", Button)

        if new_value:
            button.disabled = False
            button.variant = "success"
        else:
            button.disabled = True
            button.variant = "default"

    def watch_save_enabled(self, new_value: bool) -> None:
        button = self.query_one("#save", Button)

        if new_value:
            button.disabled = False
        else:
            button.disabled = True



from textual.app import ComposeResult
from textual.containers import Container
from textual.widget import Widget
from textual.widgets import LoadingIndicator, Static


class TxLoading(Widget):
    DEFAULT_CSS = """
    # TxLoading>Container {
    #     max-width: 16;
    #     height: 100%;
    #     opacity: 0%;
    #     align: left middle;
    #     margin-right: 2;
    # }
    # TxLoading > Vertical {
    #     max-height: 5;
    #     border: round $primary;
    # }
    # #scanning {
    #     margin-left: 2;
    #     width: auto;
    # }
    # #transactions {
    #     margin-left: 1;
    #     width: auto;
    # }
    # TxLoading LoadingIndicator {
    #     max-width: 15;
    #     max-height: 1;
    #     # height: auto;
    #     # width: auto;
    #     border: blank;
    # }
    """

    def compose(self) -> ComposeResult:
        yield Container(
            Static("Scanning", id="tx_loading_scanning"),
            LoadingIndicator(),
            Static("Transactions", id="tx_loading_transactions"),
        )



from __future__ import annotations

from typing import Iterable

from fluxwallet.keys import HDKey
from fluxwallet.mnemonic import Mnemonic as WalletMnemonic
from fluxwallet.wallet import Wallet, WalletError
from textual import on, work
from textual.app import ComposeResult, RenderResult
from textual.containers import Center, Container, Grid, Horizontal
from textual.events import Paste
from textual.message import Message
from textual.reactive import reactive, var
from textual.screen import Screen
from textual.validation import Length
from textual.widget import Widget
from textual.widgets import Button, Input, Label, Select

from lucidwallet.helpers import init_app
from lucidwallet.screens import WalletLanding

languages = [
    "english",
    "spanish",
    "italian",
    "dutch",
    "french",
    "japanese",
    "chinese_simplified",
    "chinese_traditional",
]


class MnemonicInput(Input):
    class MnemonicPaste(Message):
        def __init__(self, text: str):
            self.text = text
            super().__init__()

    def _on_paste(self, event: Paste):
        event.prevent_default()
        event.stop()
        self.post_message(self.MnemonicPaste(event.text))


class MnemonicWord(Widget, can_focus=True):
    value = reactive("")
    bip39_valid = var(False)
    wordlist = var([])

    class MnemonicWordFocused(Message):
        def __init__(self, word: MnemonicWord):
            self.word = word
            super().__init__()

    class MnemonicWordInvalidated(Message):
        def __init__(self, word: MnemonicWord):
            self.word = word
            super().__init__()

    def __init__(self, wordlist: list, *args, **kwargs):
        self._wordlist = wordlist
        super().__init__(*args, **kwargs)

    def on_mount(self):
        self.wordlist = self._wordlist

    def render(self) -> RenderResult:
        return self.value

    def on_focus(self):
        self.post_message(self.MnemonicWordFocused(self))

    def watch_value(self, new_value: str):
        if new_value in self.wordlist:
            self.bip39_valid = True
            self.add_class("-found")
        else:
            if self.bip39_valid == True:
                self.post_message(self.MnemonicWordInvalidated(self))

            self.bip39_valid = False
            self.remove_class("-found")

    def watch_wordlist(self, new_value: list):
        if self.value in new_value:
            self.bip39_valid = True
            self.add_class("-found")
        else:
            self.bip39_valid = False
            self.remove_class("-found")


class LanguagePicker(Widget):
    def __init__(self, language: str = "english"):
        self.language = language
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Label("Select Language")
        yield Select(
            [(x, x) for x in languages],
            prompt="Select language",
            value=self.language,
        )


class ImportFromMnemonic(Screen):
    DEFAULT_CSS = """
    """

    BINDINGS = [
        (
            "escape",
            "reset_and_dismiss()",
            "home",
        ),
    ]

    AUTO_FOCUS = "#nickname"

    selected_word: MnemonicWord | None = var(None)
    word_count = var(0)
    last_focused: MnemonicWord | None = var(None)
    nickname = var("")
    validated = var(False)
    submit_enabled = var(False)

    class WalletCreated(Message):
        def __init__(self, wallet: Wallet) -> None:
            self.wallet = wallet

            super().__init__()

    # class WalletLandingRequested(Message):
    #     def __init__(self, wallet: str, wallets: list[str], networks: list[str]):
    #         self.wallet = wallet
    #         self.wallets = wallets
    #         self.networks = networks
    #         super().__init__()

    def on_mount(self) -> None:
        self.wallet_names = []

    def on_screen_resume(self) -> None:
        self.store_wallet_names()

    def compose(self) -> ComposeResult:
        self.mnemonic = WalletMnemonic()

        yield LanguagePicker()
        yield Input(
            "",
            placeholder="Wallet nickname",
            id="nickname",
            validators=[Length(maximum=15)],
        )
        yield MnemonicInput(
            "", placeholder="Type mnemonic words or paste phrase", id="text"
        )
        yield Grid()
        yield Label(
            "Incorrect checksum, not a mnemonic phrase",
            id="wrong_checksum",
            classes="hidden",
        )
        yield Horizontal(
            Button("Reset Mnemonic", id="reset_button", variant="warning"),
            Button(
                "Import Wallet", id="import_button", variant="success", disabled=True
            ),
        )

    async def start_new_word(self, remove_input: bool = True) -> MnemonicWord:
        if remove_input:
            input = self.query_one("#text", Input)
            input.value = ""

        grid = self.query_one("Grid", Grid)
        self.word_count += 1
        new_word = MnemonicWord(
            self.mnemonic.wordlist(), classes="words", id=f"word_{self.word_count}"
        )
        await grid.mount(new_word)
        return new_word

    def valid_checksum(self, words: Iterable[MnemonicWord]) -> bool:
        words = " ".join([word.value for word in words])

        try:
            self.mnemonic.to_entropy(words)
        except ValueError:
            checksum = self.query_one("#wrong_checksum", Label)
            if checksum.has_class("hidden"):
                checksum.remove_class("hidden")

            return False
        else:
            return True

    def validate_words(self):
        words: Iterable[MnemonicWord] = self.query("MnemonicWord")
        import_button = self.query_one("#import_button", Button)

        if (
            len(words) == 12
            and all([x.bip39_valid for x in words])
            and self.valid_checksum(words)
        ):
            input = self.query_one("#text", Input)
            input.value = ""
            self.validated = True
            # so we can update after select does
            self.call_after_refresh(self.set_focus, import_button)

        else:
            self.validated = False
            invalid = (x for x in words if not x.bip39_valid)
            next_word = next(invalid, None)

            if next_word and next_word.value:
                self.selected_word = next_word
                self.set_focus(next_word)

    def action_focus_next(self):
        focus_chain = self.focus_chain
        if self.last_focused:
            self.last_focused = None
            self.selected_word.remove_class("-focused")
            selected = self.selected_word
        else:
            selected = self.focused

        current_index = self.focus_chain.index(selected)
        if current_index + 1 == len(focus_chain):
            self.set_focus(focus_chain[0])
        else:
            self.set_focus(focus_chain[current_index + 1])

    def action_focus_previous(self):
        focus_chain = self.focus_chain
        if self.last_focused:
            self.last_focused = None
            self.selected_word.remove_class("-focused")
            selected = self.selected_word
        else:
            selected = self.focused

        current_index = self.focus_chain.index(selected)
        if current_index == 0:
            self.set_focus(focus_chain[-1])
        else:
            self.set_focus(focus_chain[current_index - 1])

    def compute_submit_enabled(self):
        return True if self.validated and len(self.nickname) > 2 else False

    def watch_submit_enabled(self, new_value: bool):
        submit = self.query_one("#import_button")

        if new_value:
            submit.disabled = False
        else:
            submit.disabled = True

    @work(group="store_wallet_names", exclusive=True)
    async def store_wallet_names(self) -> None:
        app_data = await init_app()
        self.wallet_names = app_data.wallets

    @on(MnemonicWord.MnemonicWordFocused)
    def on_mnemonic_word_focused(self, event: MnemonicWord.MnemonicWordFocused):
        if self.selected_word:
            self.selected_word.remove_class("-focused")

        self.selected_word = event.word
        self.selected_word.add_class("-focused")

        words = self.query("MnemonicWord")
        last_word: MnemonicWord = words.last()

        if self.word_count < 12 and (not last_word.value or not last_word.bip39_valid):
            self.word_count -= 1
            last_word.remove()

        input = self.query_one("#text", Input)
        input.value = self.selected_word.value
        self.last_focused = self.selected_word
        self.set_focus(input)

    @on(MnemonicInput.MnemonicPaste)
    async def on_mnemonic_paste(self, event: MnemonicInput.MnemonicPaste) -> None:
        new_words = list(filter(None, event.text.split(" ")))

        if not self.selected_word:
            self.selected_word = await self.start_new_word()

        for index, word in enumerate(new_words, 1):
            self.selected_word.value = word

            if index < len(new_words):
                self.selected_word = await self.start_new_word()

        if len(new_words) < 12:
            self.selected_word = await self.start_new_word()

        self.validate_words()

    @on(MnemonicWord.MnemonicWordInvalidated)
    def on_mnemonic_word_invalidated(self) -> None:
        self.validated = False

    @on(Button.Pressed, "#reset_button")
    def reset_words(self) -> None:
        self.validated = False

        words: Iterable[MnemonicWord] = self.query("MnemonicWord")

        for word in words:
            word.remove()

        self.selected_word = None
        self.word_count = 0

        input = self.query_one("#text", Input)
        input.value = ""
        self.set_focus(input)

    def reset_all(self) -> None:
        self.reset_words()
        nickname = self.query_one("#nickname", Input)
        nickname.value = ""
        self.set_focus(nickname)

    @on(Input.Submitted, "#text")
    async def on_input_submitted(self, event: Input.Submitted):
        event.stop()

        if not event.value:
            return

        if self.selected_word.has_class("-focused"):
            self.selected_word.remove_class("-focused")

        if self.word_count < 12:
            self.selected_word = await self.start_new_word()
            return

        self.validate_words()

    @on(Input.Submitted, "#nickname")
    def on_nickname_submitted(self, event: Input.Submitted):
        if len(event.value) < 3:
            self.notify("Nickname must be at least 3 characters", serverity="error")

            return

        text = self.query_one("#text", Input)
        self.set_focus(text)

    def get_mnemonic_from_dom(self) -> str:
        words: Iterable[MnemonicWord] = self.query("MnemonicWord")
        return " ".join([word.value for word in words])

    @on(Button.Pressed, "#import_button")
    async def on_import(self, event: Button.Pressed):
        # only submit once
        event.button.disabled = True

        mnemonic = self.get_mnemonic_from_dom()
        # need to validate the mnemonic first, if someone just enters words willy nilly, willl get checksum error
        # ValueError: Invalid checksum 1000 for entropy b'\xc8\xcfum\x8dq9{\xafjLe\x97/\x14\x8b'
        if self.nickname in self.wallet_names:
            self.app.notify("Wallet name already exists")
            event.button.disabled = False
            return

        if await self.create_wallet(mnemonic):
            self.app.switch_screen("wallet_landing")
            self.reset_all()
        else:
            self.app.notify("Error creating Wallet")

    # @work(exclusive=True)
    async def create_wallet(self, mnemonic: str) -> bool:
        # have already validated, but validate again
        seed = self.mnemonic.to_seed(mnemonic).hex()
        hdkey = HDKey.from_seed(seed, network="flux")

        try:
            wallet = await Wallet.create(
                self.nickname,
                hdkey,
                network="flux",
            )
        except WalletError:
            return False

        if wallet:
            await wallet.new_key(network="bitcoin")
            app_data = await init_app()

            if not self.app.is_screen_installed("wallet_landing"):
                self.app.install_screen(
                    WalletLanding(
                        wallet,
                        app_data.networks,
                        app_data.wallets,
                    ),
                    name="wallet_landing",
                )

            self.post_message(self.WalletCreated(wallet))
        return True

    @on(Input.Changed, "#nickname")
    def on_nickname_changed(self, event: Input.Changed):
        if event.validation_result.failures:
            self.notify(event.validation_result.failures, serverity="error")

            return

        if event.value:
            self.nickname = event.value

    @on(Input.Changed, "#text")
    async def on_input_changed(self, event: Input.Changed):
        event.stop()

        checksum = self.query_one("#wrong_checksum", Label)
        checksum.add_class("hidden")

        new_words = list(filter(None, event.value.split(" ")))
        if len(new_words) > 1:
            return

        words: Iterable[MnemonicWord] = self.query("MnemonicWord")
        if not event.value and all([x.bip39_valid for x in words]):
            return

        if not self.selected_word:
            self.selected_word = await self.start_new_word(remove_input=False)

        if event.value.endswith(" "):
            self.selected_word.value = event.value[:-1]
            if self.word_count == 12:
                event.input.value = event.value[:-1]
                event.input.action_end()
                return

            self.selected_word.remove_class("-focused")
            self.selected_word = await self.start_new_word()
            return

        self.selected_word.value = event.value

    def on_select_changed(self, event: Select.Changed) -> None:
        event.stop()

        self.mnemonic.change_language(event.value)
        words: Iterable[MnemonicWord] = self.query("MnemonicWord")
        for word in words:
            word.wordlist = self.mnemonic.wordlist()

        self.selected_word.remove_class("-focused")
        self.validate_words()

    def action_reset_and_dismiss(self) -> None:
        self.reset_all()
        if self.app.is_screen_installed("wallet_landing"):
            self.dismiss()
        else:
            self.app.switch_screen("welcome")

    # @on(WalletLandingRequested)
    # def wallet_landing_requested(self, event: WalletLandingRequested):
    #     if not self.app.is_screen_installed("wallet_landing"):
    #         self.app.install_screen(
    #             WalletLanding(
    #                 event.wallet, event.networks, event.wallets, full_scan_required=True
    #             ),
    #             name="wallet_landing",
    #         )
    #     self.app.switch_screen("wallet_landing")
    #     self.reset_all()


# seed = Mnemonic().to_seed(passphrase).hex()
# hdkey = HDKey.from_seed(seed, network=args.network)
# return Wallet.create(
#     wallet_name,
#     hdkey,
#     network=args.network,
#     witness_type=args.witness_type,
#     db_uri=db_uri,
# )



from dataclasses import dataclass, field
from enum import Enum


@dataclass
class Event:
    class EventType(Enum):
        NewRows = "NewRows"
        ClearTable = "ClearTable"
        ScanningStart = "ScanningStart"
        ScanningEnd = "ScanningEnd"
        Scroll = "Scroll"

    type: EventType
    rows: list[tuple] = field(default_factory=list)
    scroll_height: float = 0.0



import pyperclip
from rich.text import Text
from textual import on
from textual.app import ComposeResult
from textual.containers import Container
from textual.events import Message
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import Button, Label, Select, Static


class Notification(Static):
    def on_mount(self) -> None:
        self.set_timer(3, self.remove)

    def on_click(self) -> None:
        self.remove()


class CopyLabel(Label):
    def action_copy_clipboard(self, text: str):
        try:
            self.app.config.copy_callback(text)
        except Exception:  # Fix
            self.notify("No clipboard available", severity="warning")
        else:
            self.post_message(self.Copied())

    class Copied(Message):
        ...


class TopBar(Widget):
    DEFAULT_CSS = """


    """

    balance = var(0.0)
    receive_address = var("")
    # selected_network = var("")
    # selected_wallet = var("")
    # wallets = var([])
    # networks = var([])

    class WalletSelected(Message):
        def __init__(self, wallet: str) -> None:
            self.wallet = wallet
            super().__init__()

    class NetworkSelected(Message):
        def __init__(self, network: str) -> None:
            self.network = network
            super().__init__()

    class SignMessageRequested(Message):
        ...

    class RescanAllRequested(Message):
        ...

    def __init__(
        self,
        balance: float = 0.0,
        selected_wallet: str = "",
        receive_address: str = "",
        wallets: list[str] | None = None,
        selected_network: str = "flux",
        networks: list[str] | None = None,
    ) -> None:
        self._balance = balance
        self._receive_address = receive_address

        self.selected_wallet = selected_wallet
        self.selected_network = selected_network
        self.wallets = wallets if wallets else []
        self.networks = networks if networks else []

        super().__init__()

    def on_mount(self):
        self.balance = self._balance
        self.receive_address = self._receive_address

        self.selected_wallet = self.selected_wallet
        self.selected_network = self.selected_network
        self.networks = self.networks
        self.wallets = self.wallets

        sign = self.query_one("#sign", Button)
        sign.tooltip = "Sign Message"
        rescan = self.query_one("#rescan_all", Button)
        rescan.tooltip = "Rescan all keys"

    def compose(self) -> ComposeResult:
        yield CopyLabel("", id="receive")
        yield Container(
            Label("", id="balance"),
            Select(
                id="topbar_network_id",
                options=[(x, x) for x in self.networks],
                value=self.selected_network,
            ),
        )
        yield Select(
            id="topbar_wallet_id",
            options=[(x, x) for x in self.wallets],
            value=self.selected_wallet,
            classes="wallet_select",
        )
        yield Button("\U0001F50E", id="rescan_all", classes="topbar_button")
        yield Button("\U0001F58B", id="sign", classes="topbar_button")

    def on_copy_label_copied(self):
        self.mount(Notification(Text("Copied!")))
        # self.notify("Receive address copied!", timeout=3)

    @on(Select.Changed, "#topbar_wallet_id")
    def on_wallet_changed(self, event: Select.Changed):
        event.stop()

        self.selected_wallet = event.value

        if event.value:
            self.post_message(self.WalletSelected(event.value))

    @on(Select.Changed, "#topbar_network_id")
    def on_network_changed(self, event: Select.Changed):
        event.stop()

        self.selected_network = event.value

        if event.value:
            self.post_message(self.NetworkSelected(event.value))

    def watch_balance(self, new_value: str) -> None:
        self._balance = new_value
        self.query_one("#balance", Label).update(f"Balance: { self._balance}")

    def watch_receive_address(self, new_value: str) -> None:
        self._receive_address = new_value
        receive = self.query_one("#receive", CopyLabel)
        # temporary until I sort out apis
        if self.selected_network != "flux":
            receive.update(f"Receive Address: {'ENABLING SOON!'.center(35, ' ')}")
        else:
            receive.update(
                f"Receive Address: [@click=copy_clipboard('{new_value}')]{new_value}[/]"
            )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "sign":
            self.post_message(self.SignMessageRequested())
        if event.button.id == "rescan_all":
            self.post_message(self.RescanAllRequested())

    def set_wallet_options(
        self,
        selected_wallet: str,
        selected_network: str,
        wallets: list[str],
        networks: list[str],
    ) -> None:
        self.selected_wallet = selected_wallet
        self.selected_network = selected_network
        self.wallets = wallets
        self.networks = networks

        wallet_select = self.query_one("#topbar_wallet_id", Select)
        network_select = self.query_one("#topbar_network_id", Select)

        wallet_select.set_options([(x, x) for x in wallets])
        wallet_select.value = selected_wallet

        network_select.set_options([(x, x) for x in networks])
        network_select.value = selected_network



#
#
from __future__ import annotations
from textual.app import App
from textual.widget import Widget
from textual.widgets import Static
from rich.segment import Segment
from textual.strip import Strip
from textual.geometry import Size
from rich.style import Style
from textual.containers import Container, Horizontal
from rich.color import Color
import hashlib
import importlib_metadata
from packaging import version
from time import monotonic

import asyncio

from importlib_resources import files

package = "lucidwallet"
flux_img = files(f"{package}.images").joinpath("flux_200x200.png")

import httpx
from textual import work

from PIL import Image
import numpy as np
from pathlib import Path
from numpy.typing import NDArray

from textual.screen import Screen

import keyring

from lucidwallet.helpers import init_app
from lucidwallet.screens import (
    EncryptionPassword,
    WalletLanding,
)


class Renderer:
    @classmethod
    def from_image(cls, path: Path) -> Renderer:
        img = np.asarray(Image.open(path))

        if len(img.shape) == 3:
            if img.shape[2] > 3:
                arr = np.array([[pixel[:3] for pixel in row] for row in img])
        elif len(img.shape) == 2:
            arr = np.array([[[pixel] * 3 for pixel in row] for row in img])

        return Renderer(arr)

    def __init__(self, img: NDArray) -> None:
        self.pixels = img

    @staticmethod
    def get_pixel(col: tuple[float, float, float]) -> Segment:
        color = Color.from_rgb(col[0], col[1], col[2])
        style = Style(bgcolor=color)

        return Segment("  ", style)

    def render_image(self, scale: tuple[float, float]) -> list[list[Segment]]:
        # first of all scale the image to the scale 'tuple'
        image_size = self.pixels.shape[:2]
        block_size = (image_size[0] / scale[0], image_size[1] / scale[1])
        blocks = []
        y = 0

        while y < image_size[0]:
            x = 0
            block_col = []
            while x < image_size[1]:
                # get a block, reshape in into an Nx3 matrix and then get average of each column
                block_col.append(
                    self.pixels[
                        int(y) : int(y + block_size[0]), int(x) : int(x + block_size[1])
                    ]
                    .reshape(-1, 3)
                    .mean(axis=0)
                )
                x += block_size[1]
            blocks.append(block_col)
            y += block_size[0]
        output = [[self.get_pixel(block) for block in row] for row in blocks]

        return output


class TextualImage(Widget):
    DEFAULT_CSS = """
    TextualImage {
        min-width: 60;
        min-height: 30;
        width: auto;
        height: auto;
        opacity: 0%;
    }
    """

    def __init__(self, image_path: Path) -> None:
        self.renderer = Renderer.from_image(image_path)
        self.output: list[list[Segment]] | None = None
        super().__init__()

    def on_mount(self):
        self.styles.animate("opacity", value=100.0, duration=3.0)

    def on_resize(self, new_size):
        self.generate_output()
        self.refresh()

    def generate_output(self) -> None:
        if self.size.width == self.app.size.width:
            max_size = min(self.app.size.width / 2, self.app.size.height)
            max_size = max_size * 0.8
        else:
            max_size = min(self.size.width / 2, self.size.height)

        self.output = self.renderer.render_image(scale=(max_size, max_size))

    def render_line(self, y: int) -> Strip:
        if not self.output:
            self.generate_output()

        if y + 1 > len(self.output):
            return Strip.blank(self.size.width, self.rich_style)

        row = self.output[y]

        strip = Strip(row)
        return strip


class LoadingScreen(Screen):
    DEFAULT_CSS = """
    LoadingScreen {
        width: 100%;
        height: 1fr;
        align: center middle;
        background: black;
        border: tall $primary;
    }
    LoadingScreen Static {
        border: greenyellow;
        width: auto;
    }
    """

    def compose(self):
        yield Static("Lucidwallet")
        yield TextualImage(flux_img)

    @work(name="version_check")
    async def version_check(self) -> None:
        if new_version := await self.new_version_available():
            self.call_after_refresh(
                self.notify, f"New version {new_version} available", timeout=5
            )

    async def new_version_available(self) -> str | None:
        current_version = importlib_metadata.version(package)
        current_version = version.parse(current_version)
        # shame we have to download the entire package info for just the version
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(f"https://pypi.org/pypi/{package}/json")
        except httpx.HTTPError:
            return

        try:
            latest_version = response.json()["info"]["version"]
        except KeyError:
            return

        latest_version = version.parse(latest_version)

        return latest_version if latest_version > current_version else None

    async def valid_password(self) -> bool:
        return await self.app.config.last_used_wallet.db.validate_key()

    async def encryption_password_callback(self, result: tuple[str, bool]) -> None:
        password, store_in_keychain = result

        hashed = hashlib.sha256(bytes(password, "utf8")).hexdigest()
        self.app.config.last_used_wallet.db.set_encrypted_key(hashed)

        if not await self.valid_password():
            self.app.push_screen(
                EncryptionPassword(message="Invalid Password"),
                self.encryption_password_callback,
            )
            return

        if store_in_keychain:
            keyring.set_password("fluxwallet", "fluxwallet_user", hashed)

        self.app.install_screen(
            WalletLanding(
                self.app.config.last_used_wallet,
                self.app.config.networks,
                self.app.config.wallets,
            ),
            name="wallet_landing",
        )
        self.app.switch_screen("wallet_landing")

    async def wait_loading(self) -> None:
        # slow loading for minimum 2 seconds
        elapsed = monotonic() - self.start

        if elapsed < 2:
            await asyncio.sleep(2 - elapsed)

    @work(group="boot")
    async def boot(self) -> None:
        self.app.config = await init_app()

        if not self.app.config.last_used_wallet:
            await self.wait_loading()
            self.app.switch_screen("welcome")
            return

        if self.app.config.encrypted_db:
            print("DB ENCRYPTED")
            password_hash = ""

            if self.app.config.keyring_available:
                password_hash = keyring.get_password("fluxwallet", "fluxwallet_user")

            if not password_hash:
                screen = EncryptionPassword()

                await self.wait_loading()

                self.app.push_screen(
                    screen,
                    self.encryption_password_callback,
                )
                return
            else:
                # this only ever needs to get set once
                self.app.config.last_used_wallet.db.set_encrypted_key(password_hash)

                if not await self.valid_password():
                    if self.app.config.keyring_available:
                        keyring.delete_password("fluxwallet", "fluxwallet_user")

                    screen = EncryptionPassword(message="Invalid Password")

                    await self.wait_loading()

                    self.app.push_screen(
                        screen,
                        self.encryption_password_callback,
                    )
                    return

        self.app.install_screen(
            WalletLanding(
                self.app.config.last_used_wallet,
                self.app.config.networks,
                self.app.config.wallets,
            ),
            name="wallet_landing",
        )

        await self.wait_loading()

        self.app.switch_screen("wallet_landing")

    async def on_show(self) -> None:
        self.start = monotonic()
        self.call_after_refresh(self.boot)


# if __name__ == "__main__":

#     class BlahApp(App):
#         def on_mount(self):
#             self.push_screen(LoadingScreen())

#     app = BlahApp()
#     app.run()



import hashlib
import os

import keyring
from packaging import version

from textual import on, work
from textual.app import App
from textual.binding import Binding

import asyncio
from lucidwallet.helpers import init_app
from lucidwallet.screens import (
    LoadingScreen,
    CreateWallet,
    EncryptionPassword,
    FirstRun,
    ImportFromMnemonic,
    MnemonicOverlay,
    WalletLanding,
)

import httpx
import importlib_metadata


package = "lucidwallet"


class LucidWallet(App[None]):
    CSS_PATH = "app.css"
    SCREENS = {
        "welcome": FirstRun(),
        "create_wallet": CreateWallet(),
        "from_mnemonic": ImportFromMnemonic(),
        "loading": LoadingScreen(),
    }
    BINDINGS = [
        ("ctrl+t", "app.toggle_dark", "Toggle Dark mode"),
        Binding("ctrl+c,ctrl+q", "app.quit", "Quit", show=True),
    ]

    async def new_version_available(self) -> str | None:
        current_version = importlib_metadata.version(package)
        current_version = version.parse(current_version)
        # shame we have to download the entire package info for just the version
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(f"https://pypi.org/pypi/{package}/json")
        except httpx.HTTPError:
            return

        try:
            latest_version = response.json()["info"]["version"]
        except KeyError:
            return

        latest_version = version.parse(latest_version)

        return latest_version if latest_version > current_version else None

    @work(name="version_check")
    async def version_check(self) -> None:
        if new_version := await self.new_version_available():
            self.call_after_refresh(
                self.notify, f"New version {new_version} available", timeout=5
            )

    async def valid_password(self) -> bool:
        return await self.config.last_used_wallet.db.validate_key()

    async def encryption_password_callback(self, result: tuple[str, bool]) -> None:
        password, store_in_keychain = result

        hashed = hashlib.sha256(bytes(password, "utf8")).hexdigest()
        self.config.last_used_wallet.db.set_encrypted_key(hashed)

        if not await self.valid_password():
            self.push_screen(
                EncryptionPassword(message="Invalid Password"),
                self.encryption_password_callback,
            )
            return

        if store_in_keychain:
            keyring.set_password("fluxwallet", "fluxwallet_user", hashed)

        self.install_screen(
            WalletLanding(
                self.config.last_used_wallet, self.config.networks, self.config.wallets
            ),
            name="wallet_landing",
        )
        self.push_screen("wallet_landing")

    async def boot(self) -> None:
        self.version_check()

        self.config = await init_app()

        if not self.config.last_used_wallet:
            self.push_screen("welcome")
            return

        if self.config.encrypted_db:
            password_hash = ""

            if self.config.keyring_available:
                password_hash = keyring.get_password("fluxwallet", "fluxwallet_user")

            if not password_hash:
                self.push_screen(
                    EncryptionPassword(), self.encryption_password_callback
                )
                return
            else:
                # this only ever needs to get set once
                self.config.last_used_wallet.db.set_encrypted_key(password_hash)

                if not await self.valid_password():
                    if self.config.keyring_available:
                        keyring.delete_password("fluxwallet", "fluxwallet_user")

                    self.push_screen(
                        EncryptionPassword(message="Invalid Password"),
                        self.encryption_password_callback,
                    )
                    return

        self.install_screen(
            WalletLanding(
                self.config.last_used_wallet, self.config.networks, self.config.wallets
            ),
            name="wallet_landing",
        )
        self.push_screen("wallet_landing")

    # async def on_load(self) -> None:
    #     self.push_screen("loading")

    async def on_mount(self) -> None:
        # await self.boot()
        # push loading screen first, with logo
        self.push_screen("loading")

    # fix these
    @on(MnemonicOverlay.WalletCreated)
    async def on_mnemonic_overlay_wallet_created(
        self, event: MnemonicOverlay.WalletCreated
    ):
        wallet_landing: WalletLanding = self.get_screen("wallet_landing")
        await wallet_landing.new_wallet_created(event.wallet)

    @on(ImportFromMnemonic.WalletCreated)
    async def on_import_from_mnemonic_wallet_created(
        self, event: ImportFromMnemonic.WalletCreated
    ):
        wallet_landing: WalletLanding = self.get_screen("wallet_landing")
        await wallet_landing.new_wallet_created(event.wallet)


# for console script
def run():
    # ubuntu by default doesn't have truecolor set, so the colors are weird
    os.environ["COLORTERM"] = "truecolor"
    # just in case it's been modified
    os.environ["TERM"] = "xterm-256color"

    app = LucidWallet()
    app.run()


# for textual console
if __name__ == "__main__":
    run()



from functools import partial

from fluxwallet.mnemonic import Mnemonic
from textual.app import ComposeResult
from textual.containers import Grid, Horizontal
from textual.message import Message
from textual.reactive import var
from textual.screen import Screen
from textual.validation import Length
from textual.widget import Widget
from textual.widgets import Button, Input, Label, Select, Static
from textual import work


from lucidwallet.helpers import init_app
from lucidwallet.screens import MnemonicOverlay

languages = [
    "english",
    "spanish",
    "italian",
    "dutch",
    "french",
    "japanese",
    "chinese_simplified",
    "chinese_traditional",
]


# class Notification(Static):
#     def on_mount(self) -> None:
#         self.set_timer(3, self.remove)

#     def on_click(self) -> None:
#         self.remove()


class LanguagePicker(Widget):
    def __init__(self, language: str = "english"):
        self.language = language
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Label("Select Language")
        yield Select(
            [(x, x) for x in languages],
            prompt="Select language",
            value=self.language,
        )

    def reset(self) -> None:
        self.language = "english"
        select = self.query_one("Select", Select)
        select.value = "english"

    def on_select_changed(self, event: Select.Changed):
        self.language = event.value


class CreateWallet(Screen):
    BINDINGS = [
        (
            "escape",
            "reset_and_dismiss()",
            "home",
        ),
    ]

    nickname = var("")

    # class WalletLandingRequested(Message):
    #     def __init__(self, wallet: str, wallets: list[str], networks: list[str]):
    #         self.wallet = wallet
    #         self.wallets = wallets
    #         self.networks = networks
    #         super().__init__()

    def compose(self):
        yield Static("Create Wallet", id="create_wallet_title")
        yield LanguagePicker()
        yield Input(
            "",
            placeholder="Wallet nickname",
            id="nickname",
            validators=[Length(maximum=15, minimum=2)],
        )
        yield Grid(id="mnemonic_grid")
        yield Horizontal(
            Button("Cancel", variant="warning", id="wallet_create_cancel"),
            Button("Create", id="wallet_create", disabled=True),
            id="create_wallet_button_container",
        )

    def on_mount(self) -> None:
        self.wallet_names = []

        input = self.query_one("Input", Input)
        input.focus()

    def on_screen_resume(self) -> None:
        self.store_wallet_names()

    def on_input_changed(self, event: Input.Changed) -> None:
        if event.validation_result.failures:
            # self.mount(
            #     Notification(event.validation_result.failures, variant="failure")
            # )
            return

        if event.value:
            self.nickname = event.value

    def reset_all(self) -> None:
        self.nickname = ""
        input = self.query_one("Input", Input)
        input.value = ""
        lang_picker = self.query_one("LanguagePicker", LanguagePicker)
        lang_picker.reset()
        input.focus()

    @work(group="store_wallet_names", exclusive=True)
    async def store_wallet_names(self) -> None:
        app_data = await init_app()
        self.wallet_names = app_data.wallets

    async def on_button_pressed(self, event: Button.Pressed):
        event.stop()
        if event.button.id == "wallet_create":
            print("WALLET CREATE")
            if self.nickname in self.wallet_names:
                self.notify("Wallet name already exists")
                return

            lang_picker = self.query_one("LanguagePicker", LanguagePicker)
            mnemonic = Mnemonic(lang_picker.language).generate()

            self.app.push_screen(MnemonicOverlay(self.nickname, mnemonic))
            self.reset_all()
        else:
            self.action_reset_and_dismiss()

    def action_reset_and_dismiss(self) -> None:
        self.reset_all()
        if self.app.is_screen_installed("wallet_landing"):
            self.dismiss()
        else:
            self.app.switch_screen("welcome")

    def watch_nickname(self, new_value) -> None:
        create = self.query_one("#wallet_create", Button)

        if new_value:
            create.disabled = False
        else:
            create.disabled = True



import datetime
import random
import secrets
import string

from rich.text import Text


def random_date() -> str:
    start_date = datetime.date(2022, 6, 1)
    end_date = datetime.date(2023, 6, 1)

    num_days = (end_date - start_date).days
    rand_days = random.randint(1, num_days)
    random_date = start_date + datetime.timedelta(days=rand_days)
    return str(random_date)


def random_dir() -> Text:
    if random.randint(0, 1):
        return Text("-->", style="bold green", justify="right")
    else:
        return Text("<--", style="bold red", justify="left")


def random_address() -> str:
    alphabet = string.ascii_uppercase + string.ascii_lowercase + string.digits
    return f"t1{''.join(secrets.choice(alphabet) for i in range(33))}"


def random_txid() -> str:
    alphabet = string.ascii_lowercase + string.digits
    return "".join(secrets.choice(alphabet) for i in range(64))


def random_amount() -> float:
    places = random.randint(0, 5)
    return round(random.uniform(0.0, 1000.0), places)


ROWS = []
for n in range(50):
    ROWS.append(
        (random_txid(), random_date(), random_dir(), random_address(), random_amount())
    )



from dataclasses import dataclass, field

from fluxwallet.db_new import Db, DbConfig, DbWallet
from fluxwallet.wallet import Wallet
from sqlalchemy import select
from textual.widgets import Static

import asyncio

from collections.abc import Callable

import keyring
from keyring.errors import NoKeyringError
import secrets
import pyperclip

import base64
import httpx


@dataclass
class NetworkData:
    price: float
    marketcap: int
    delta_24hr: float
    hashrate: int
    blockheight: int


@dataclass
class InitAppResponse:
    last_used_wallet: Wallet | None = None
    wallets: list[str] = field(default_factory=list)
    encrypted_db: bool = False
    networks: list[str] = field(default_factory=list)
    keyring_available: bool = False
    copy_callback: Callable | None = None
    show_market_data: bool = True
    network_data: NetworkData | None = None


async def get_db_info() -> tuple[Wallet | None, list[str], bool]:
    db = await Db.start()

    async with db as session:
        last_used_wallet: Wallet | None = None
        res = await session.scalars(select(DbWallet).order_by(DbWallet.id))
        wallets = res.all()

        if wallets:
            res = await session.scalars(
                select(DbConfig.value).filter_by(variable="last_used_wallet")
            )
            last_used_wallet_name = res.first()

            if not last_used_wallet_name:
                last_used_wallet_name = wallets[0].name
                await session.merge(
                    DbConfig(variable="last_used_wallet", value=last_used_wallet_name)
                )
                await session.commit()

            if last_used_wallet_name:
                last_used_wallet = await Wallet.open(last_used_wallet_name)

            wallets = [x.name for x in wallets]

            res = await session.scalars(
                select(DbConfig.value).filter_by(variable="show_market_data")
            )
            show_market_data = res.first()

            # if not set, we default to True
            if show_market_data is None:
                show_market_data = True
            else:
                show_market_data = bool(int(show_market_data))

        else:
            wallets = []
            show_market_data = True

    return last_used_wallet, wallets, db.encrypted, show_market_data


def write_tty(data: bytes) -> None:
    with open("/dev/tty", "wb") as f:
        f.write(data)
        f.flush()


def osc52_copy(data: str) -> None:
    data = bytes(data, encoding="utf-8")
    encoded = base64.b64encode(data)
    buffer = b"\033]52;p;" + encoded + b"\a"

    write_tty(buffer)


async def get_network_data() -> NetworkData:
    async with httpx.AsyncClient() as client:
        tasks = []
        tasks.append(
            client.get("https://api.coinpaprika.com/v1/tickers/zel-zelcash?quotes=USD")
        )

        tasks.append(
            client.get(
                "https://explorer.runonflux.io/api/status",
                params={"q": "getMiningInfo"},
            )
        )
        tasks.append(client.get("https://explorer.runonflux.io/api/sync"))
        market_data, mining_data, sync_info = await asyncio.gather(*tasks)
        market_data = market_data.json()
        mining_data = mining_data.json()
        sync_info = sync_info.json()

        price = market_data["quotes"]["USD"]["price"]
        marketcap = market_data["quotes"]["USD"]["market_cap"]
        delta_24hr = market_data["quotes"]["USD"]["percent_change_24h"]
        hashrate = mining_data["miningInfo"]["networkhashps"]
        blockheight = sync_info["height"]

        return NetworkData(price, marketcap, delta_24hr, hashrate, blockheight)


async def init_app() -> InitAppResponse:
    network_task = asyncio.create_task(get_network_data())

    (
        last_used_wallet,
        known_wallets,
        db_encrypted,
        show_market_data,
    ) = await get_db_info()

    random_user = secrets.token_hex(8)

    keyring_available = True
    try:
        # can use empty strings but seems dodgey
        keyring.get_password(random_user, random_user)
    except NoKeyringError:
        keyring_available = False

    try:
        # this will remove any copied stuff though
        pyperclip.copy("")
    except pyperclip.PyperclipException:
        copy_callback = osc52_copy
    else:
        copy_callback = pyperclip.copy

    wallet_networks = []
    if last_used_wallet:
        # this coudl error
        keys = await last_used_wallet.keys_networks()
        wallet_networks = [x.network.name for x in keys]

    network_data = await network_task

    return InitAppResponse(
        last_used_wallet,
        known_wallets,
        db_encrypted,
        wallet_networks,
        keyring_available,
        copy_callback,
        show_market_data,
        network_data,
    )



import webbrowser

from fluxwallet.transactions import BaseTransaction
from fluxwallet.wallet import WalletTransaction
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container, Grid, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class KeychainOverlay(ModalScreen[bool]):
    DEFAULT_CSS = """
        KeychainOverlay {
            align: center middle;
        }

        KeychainOverlay #dialog {
            grid-size: 2;
            grid-gutter: 1 2;
            grid-rows: 1fr 3;
            padding: 0 1;
            width: 60;
            height: 11;
            border: thick $background 80%;
            background: $surface;
        }

        KeychainOverlay #question {
            column-span: 2;
            height: 1fr;
            width: 1fr;
            content-align: center middle;
        }

        KeychainOverlay Button {
            width: 100%;
        }
    """

    BINDINGS = [
        ("ctrl+t", "app.toggle_dark", "Toggle Dark mode"),
        Binding("ctrl+c,ctrl+q", "app.quit", "Quit"),
    ]

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(
                "Do you want to store password in Keychain if available?", id="question"
            ),
            Button("Yes", variant="error", id="yes"),
            Button("No", variant="primary", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "no":
            self.dismiss(False)
        else:
            self.dismiss(True)



import asyncio
import time
from collections.abc import Awaitable
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum

from fluxwallet.db_new import DbKey
from fluxwallet.values import Value
from fluxwallet.wallet import Wallet, WalletTransaction
from rich.text import Text

from lucidwallet.events import Event


class ScanType(Enum):
    PERIODIC = "PERIODIC"
    NEW_KEY = "NEW_KEY"
    FULL_WALLET = "FULL_WALLET"


@dataclass
class Timer:
    callback: Awaitable
    interval: int = 10
    run_on_start: bool = True
    loop_task: asyncio.Task | None = None

    def start(self) -> None:
        if not self.loop_task or self.loop_task.cancelled():
            self.loop_task = asyncio.create_task(self.loop())

    async def reset(self, run_on_start: bool = True) -> None:
        self.run_on_start = run_on_start

        await self.stop()
        self.start()

    async def stop(self) -> None:
        if not self.loop_task:
            return

        self.loop_task.cancel()

        try:
            await self.loop_task
        except asyncio.CancelledError:
            pass

    async def loop(self) -> None:
        if self.run_on_start:
            await self.callback()

        while True:
            await asyncio.sleep(self.interval)
            await self.callback()


@dataclass
class Network:
    offset: int = 0
    transactions: list[WalletTransaction] = field(default_factory=list)
    table_rows: list[tuple] = field(default_factory=list)
    last_scanned_periodic: float = 0.0
    last_scanned_full: float = 0.0
    scroll_height: float = 0.0
    scanning: bool = False
    first_scan: bool = True
    balance: float = 0.0  # cache value for when scanning

    def reset(self) -> None:
        """Default network back to starting values"""
        self.offset = 0
        self.last_scanned = 0.0
        self.scanning = False
        self.transactions = []
        self.table_rows = []
        self.scroll_height = 0.0
        self.first_scan = True


@dataclass
class WalletData:
    wallet: Wallet
    networks: dict[str, Network] = field(default_factory=dict)

    def reset_networks(self) -> None:
        """Reset all networks back to default values"""
        for network in self.networks.values():
            network.reset()


@dataclass
class FluxWalletDataStore:
    current_wallet: str | None = None
    current_network: str = "flux"
    wallet_names: set[str] = field(default_factory=set)
    wallets: dict[str, WalletData] = field(default_factory=dict)
    unconfirmed_txs: list[WalletTransaction] = field(default_factory=list)
    scan_timer: Timer | None = None
    tx_history_scanning: bool = False

    def get_current_wallet(self) -> Wallet:
        return self.wallets[self.current_wallet].wallet

    def get_current_wallet_networks(self) -> list[str]:
        return self.wallets[self.current_wallet].networks

    def get_transaction(self, txid: str) -> WalletTransaction | None:
        wallet_data = self.wallets[self.current_wallet]

        tx = next(
            filter(
                lambda x: x.tx.txid == txid,
                wallet_data.networks[self.current_network].transactions,
            ),
            None,
        )

        return tx

    def add_timer(self, timer: Timer):
        self.scan_timer = timer

    def exists(self, wallet: str | Wallet) -> bool:
        if isinstance(wallet, Wallet):
            wallet = wallet.name

        return True if wallet in self.wallets else False

    async def store_wallet(self, wallet: str | Wallet) -> None:
        if isinstance(wallet, str):
            wallet = await Wallet.open(wallet)

        keys = await wallet.keys_networks()
        networks = {x.network.name: Network() for x in keys}
        self.wallets[wallet.name] = WalletData(wallet, networks)

    async def receive_address(self, wallet_name: str, network: str) -> str:
        next_unused_key = await self.wallets[wallet_name].wallet.get_key(
            network=network
        )
        return next_unused_key.address

    def get_last_network_balance(self, wallet_name: str, network_name: str) -> float:
        return self.wallets[wallet_name].networks[network_name].balance

    def set_last_network_balance(
        self, wallet_name: str, network_name: str, balance: float
    ) -> float:
        self.wallets[wallet_name].networks[network_name].balance = balance

    async def balance(self, wallet_name: str, network: str) -> float:
        if self.is_network_scanning(wallet_name, network):
            balance = self.get_last_network_balance(wallet_name, network)
        else:
            balance = await self.wallets[wallet_name].wallet.balance(network=network)
            self.set_last_network_balance(wallet_name, network, balance)

        return balance

    async def current_network_balance(self) -> float:
        return await self.balance(self.current_wallet, self.current_network)

    async def set_current_wallet(self, wallet: str | Wallet) -> None:
        if not self.exists(wallet):
            await self.store_wallet(wallet)

        self.current_wallet = str(wallet)

        # change this to default or something
        self.set_current_network("flux")

    def set_current_network(self, network_name: str) -> None:
        self.current_network = network_name

    def get_current_network(self) -> str:
        return self.current_network

    async def get_current_wallet_spend_info(self) -> tuple[str, float]:
        receive_address = await self.receive_address(
            self.current_wallet, self.current_network
        )

        balance = await self.balance(self.current_wallet, self.current_network)

        return (receive_address, balance)

    async def parse_wallet_txs(
        self, wallet_name: str, wallet_txs: list[WalletTransaction]
    ) -> tuple:
        TX_DIRECTION = {
            False: Text("-->", style="bold green", justify="right"),
            True: Text("<--", style="bold red", justify="left"),
        }

        parsed_txs = []

        used_keys: list[DbKey] = await self.wallets[wallet_name].wallet.keys(
            used=True, network=self.current_network
        )
        used_addresses = [x.address for x in used_keys]

        for wt in wallet_txs:
            if wt.tx.date:
                date = wt.tx.date
                date = date.replace(tzinfo=timezone.utc)
                # local
                date = date.astimezone()
            else:
                date = datetime.now()

            outbound = False
            for input in wt.tx.inputs:
                if input.address in used_addresses:
                    outbound = True
                    break

            value = 0
            for output in wt.tx.outputs:
                if outbound:
                    if not output.address in used_addresses:
                        # this is who the output is to
                        value += output.value
                else:
                    if output.address in used_addresses:
                        value += output.value

            value = Value.from_satoshi(value, network=self.current_network).str(
                1, currency_repr=None
            )
            parsed_txs.append(
                (
                    wt.tx.txid,
                    TX_DIRECTION[outbound],
                    date.date(),
                    date.strftime("%H:%M:%S"),
                    value,
                    self.current_network.upper(),
                )
            )

        return parsed_txs

    async def get_txs(
        self,
        response: asyncio.Queue,
        wallet_name: str,
        network: str,
        latest_only: bool = False,
        limit: int = 60,
    ) -> None:
        offset = (
            0 if latest_only else self.wallets[wallet_name].networks[network].offset
        )

        tx_gen = self.wallets[wallet_name].wallet.transactions_full(
            network=network, limit=limit, offset=offset
        )

        async for txs in tx_gen:
            self.wallets[wallet_name].networks[network].offset += len(txs)

            self.wallets[wallet_name].networks[network].transactions.extend(txs)

            if parsed_txs := await self.parse_wallet_txs(wallet_name, txs):
                self.wallets[wallet_name].networks[network].table_rows.extend(
                    parsed_txs
                )
                await response.put(Event(type=Event.EventType.NewRows, rows=parsed_txs))

    async def get_transactions_from_db(
        self,
        response: asyncio.Queue,
        force: bool = False,
        latest_only: bool = False,
        limit: int = 60,
    ) -> None:
        """_summary_

        Args:
            force (bool, optional): force db query for new txs. Defaults to False.
            latest_only (bool, optional): will force db query and only retrieve latest
            txs from db, ignoring internal offset. Defaults to False.
            limit (int, optional): how many txs to fetch. Defaults to 60.
        """
        if not force and self.current_network_has_table_rows():
            await response.put(
                Event(
                    type=Event.EventType.NewRows, rows=self.current_network_table_rows()
                )
            )
            await response.put(
                Event(
                    type=Event.EventType.Scroll,
                    scroll_height=self.get_current_network_scroll_height(),
                )
            )
            return

        await self.get_txs(
            response,
            self.current_wallet,
            self.current_network,
            latest_only,
            limit=limit,
        )

    def current_network_has_table_rows(self) -> bool:
        has_rows = bool(len(self.current_network_table_rows()))
        return has_rows

    def current_network_table_rows(self) -> list[tuple]:
        return (
            self.wallets[self.current_wallet].networks[self.current_network].table_rows
        )

    def store_current_network_scroll_height(self, scroll_height: int) -> None:
        self.wallets[self.current_wallet].networks[
            self.current_network
        ].scroll_height = scroll_height

    def get_current_network_scroll_height(self) -> int:
        return (
            self.wallets[self.current_wallet]
            .networks[self.current_network]
            .scroll_height
        )

    def reset_current_wallet_datastore(self) -> None:
        self.wallets[self.current_wallet].reset_networks()

    def set_scanning_for_network(self, wallet_name: str, network_name: str) -> None:
        self.wallets[wallet_name].networks[network_name].scanning = True

    def unset_scanning_for_network(self, wallet_name: str, network_name: str) -> None:
        self.wallets[wallet_name].networks[network_name].scanning = False

    def is_network_scanning(self, wallet_name: str, network_name: str) -> bool:
        return self.wallets[wallet_name].networks[network_name].scanning

    def is_current_network_scanning(self) -> bool:
        return self.is_network_scanning(self.current_wallet, self.current_network)

    def set_last_scanned_for_network(
        self, wallet: str, network: str, scan_type: ScanType
    ) -> None:
        if scan_type == ScanType.PERIODIC:
            self.wallets[wallet].networks[
                network
            ].last_scanned_periodic = time.monotonic()
        else:  # Full
            self.wallets[wallet].networks[network].last_scanned_full = time.monotonic()

    def scan_for_current_network_required(self, scan_type: ScanType) -> bool:
        scan_required = True
        now = time.monotonic()

        recent_periodic = (
            self.get_current_network_last_scan_time(ScanType.PERIODIC) + 15 > now
        )
        recent_full = (
            self.get_current_network_last_scan_time(ScanType.FULL_WALLET) + 15 > now
        )

        if scan_type == ScanType.FULL_WALLET:
            if recent_full:
                scan_required = False

        if scan_type == ScanType.PERIODIC:
            if recent_periodic or recent_full:
                scan_required = False

        return scan_required

    def get_current_network_last_scan_time(self, scan_type: ScanType) -> float:
        # fix all this
        if scan_type == ScanType.PERIODIC:
            return (
                self.wallets[self.current_wallet]
                .networks[self.current_network]
                .last_scanned_periodic
            )
        else:  # full
            return (
                self.wallets[self.current_wallet]
                .networks[self.current_network]
                .last_scanned_full
            )

    def is_network_first_scan(self, wallet_name: str, network_name: str) -> bool:
        return self.wallets[wallet_name].networks[network_name].first_scan

    def set_network_scanned(self, wallet_name: str, network_name: str) -> None:
        self.wallets[wallet_name].networks[network_name].first_scan = False

    def add_known_wallet(self, wallet: str) -> None:
        self.wallet_names.add(wallet)

    def get_known_wallets(self) -> list:
        return list(self.wallet_names)

    def add_unconfirmed_tx(self, tx: WalletTransaction) -> None:
        self.unconfirmed_txs.append(tx)

    def start_scan_timer(self) -> None:
        if self.scan_timer:
            self.scan_timer.start()

    async def reset_timer(self, run_on_start: bool = True) -> None:
        await self.scan_timer.reset(run_on_start)
